/***************************************************************************
 *   Copyright (C) 2007 by Lisandro Damián Nicanor Pérez Meyer             *
 *   perezmeyer@gmail.com                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "graphicsscene.h"
#include "point.h"
#include <math.h>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsLineItem>
#include <QLineF>
#include <QVector>
#include <QString>
#include <QDebug>

GraphicsScene::GraphicsScene(QObject * parent) : QGraphicsScene(parent)
{
	lagrangeCurve = false;
	hermiteCurve = false;
	deCasteljauCurve = false;
	bernsteinCurve = false;
	bSplinesCurve = false;
	cubicBSplineCurve = false;
	bSplineGrade = 1;
}

void GraphicsScene::displayLagrangeCurve(int state)
{
	if(state == Qt::Checked)
		lagrangeCurve = true;
	else
		lagrangeCurve = false;
	drawLagrange();
}

void GraphicsScene::displayHermiteCurve(int state)
{
	if(state == Qt::Checked)
		hermiteCurve = true;
	else
		hermiteCurve = false;
	drawHermite();
}

void GraphicsScene::displayDeCasteljauCurve(int state)
{
	if(state == Qt::Checked)
		deCasteljauCurve = true;
	else
		deCasteljauCurve = false;
	drawDeCasteljau();
}

void GraphicsScene::displayBernsteinCurve(int state)
{
	if(state == Qt::Checked)
		bernsteinCurve = true;
	else
		bernsteinCurve = false;
	drawBernstein();
}

void GraphicsScene::displayBSplinesCurve(int state)
{
	if(state == Qt::Checked)
		bSplinesCurve = true;
	else
		bSplinesCurve = false;
	drawBSplines();
}

void GraphicsScene::displayCubicBSplineCurve(int state)
{
	if(state == Qt::Checked)
		cubicBSplineCurve = true;
	else
		cubicBSplineCurve = false;
	drawCubicBSpline();
}

void GraphicsScene::bSplineCurveGradeChanged(int value)
{
	bSplineGrade = value;
	drawBSplines();
}

void GraphicsScene::mousePressEvent(QGraphicsSceneMouseEvent * mouseEvent)
{
	/*
		Clear the status bar. If it is need, each specific curve method will
		print a message again.
	*/
	emit sendSatusBarMessage(tr(""), 1);

	// Add a point
	if (mouseEvent->button() == Qt::LeftButton)
	{
		Point * item;
		item = new Point();
		// We need an ordered list of points, so...
		pointsList.append(item);
		// Set the correct number
		item->setNumber(pointsList.size());

		addItem(item);
		item->setPos(mouseEvent->scenePos());
	}
	// Delete a point
	else if(mouseEvent->button() == Qt::RightButton)
	{
		QGraphicsItem * item;
		item = itemAt(mouseEvent->scenePos());

		if(item != 0)
		{
			removeItem(item);
			// Remove the item from the ordered list
			for(int i=0; i< pointsList.size(); i++)
			{
				if(pointsList[i] == item)
				{
					pointsList.removeAt(i);
					break;
				}
			}
			// We created it, we delete it
			delete item;
		}
	}
	drawCurves();
}

void GraphicsScene::drawCurves()
{
	drawLagrange();
	drawHermite();
	drawDeCasteljau();
	drawBernstein();
	drawBSplines();
	drawCubicBSpline();
}

void GraphicsScene::drawLagrange()
{
	// Remove the previous lines from the scene
	for(int i=0; i< lagrangeLines.size(); i++)
		removeItem(lagrangeLines[i]);

	// Then, delete the lines
	qDeleteAll(lagrangeLines.begin(), lagrangeLines.end());
 	lagrangeLines.clear();

	// Check wheter we have to render the curve, else return
	if(!lagrangeCurve)
		return;

	// Check wheter we can draw a curve, else return
	if(pointsList.size() < 2)
		return;

	/*
		The Lagrange algorythm needs the points to be in order.
		We check that.
	*/
	for(int i=1; i < pointsList.size(); i++)
	{
		if(pointsList[i-1]->pos().x() > pointsList[i]->pos().x())
		{
			emit sendSatusBarMessage(tr("The Lagrange algorythm needs the points to "
			                            "be ordered by the x coordinate."), 5000);
			return;
		}
	}

	QGraphicsLineItem * item;

	double deltaX = pointsList.last()->pos().x() - pointsList.first()->pos().x();
	deltaX /= lagrangeStep;

	double x(0.0), y(0.0);
	double num(1.0), den(1.0);
	double lastX = pointsList.first()->pos().x();
	double lastY = pointsList.first()->pos().y();

	for(int i=1; i <= lagrangeStep; i++)
	{
		x = pointsList.first()->pos().x() + deltaX * i;
		y = 0;

		for(int j=1; j < pointsList.size(); j++)
		{
			num = 1.0;
			den = 1.0;
			for(int k=1; k < pointsList.size(); k++)
			{
				if( j != k )
				{
					num = num * (x - pointsList[k]->pos().x());
					den = den * (pointsList[j]->pos().x() - pointsList[k]->pos().x());
				}
			}
			y = y + num/den * pointsList[j]->pos().y();
		}
		item = new QGraphicsLineItem(lastX, lastY, x, y);
		item->setPen(QPen(QColor(Qt::red)));
		lagrangeLines.append(item);
		addItem(item);
		lastX = x;
		lastY = y;
	}
}

void GraphicsScene::drawHermite()
{
	// Remove the previous lines from the scene
	for(int i=0; i< hermiteLines.size(); i++)
		removeItem(hermiteLines[i]);

	// Then, delete the lines
	qDeleteAll(hermiteLines.begin(), hermiteLines.end());
 	hermiteLines.clear();

	// Check wheter we have to render the curve, else return
	if(!hermiteCurve)
		return;

	// Check wheter we can draw a curve, else return
	if(pointsList.size() < 2)
		return;

	QGraphicsLineItem * item;
	QPointF d1(0.0, 0.0), d2(0.0, 0.0);
	QPointF c0(0.0, 0.0), c1(0.0, 0.0), c2(0.0, 0.0), c3(0.0, 0.0);
	QPointF p(0.0, 0.0);
	double u = 0.0;

	for(int i=1; i < pointsList.size(); i++)
	{
		qDebug() << i;
		// Calculate d1 and d2. We assume the tangents to be 1/3 (delta points)
		if(i==1)
		{
			d1 = 1/3*(pointsList[i]->pos() - pointsList[i-1]->pos());
			d2 = -d1;
			qDebug() << "primero";
		}
		else if(i == pointsList.size() -1)
		{
			d2 = 1/3*(pointsList[i-1]->pos() - pointsList[i]->pos());
			d1 = -d2;
			qDebug() << "último";
		}
		else
		{
			d1 = 1/3*(pointsList[i+1]->pos() - pointsList[i-1]->pos());
			d2 = -d1;
			qDebug() << "el resto";
		}

		c3 = 2 * pointsList[i-1]->pos() - 2 * pointsList[i]->pos() + d2;
		c2 = -3 * pointsList[i-1]->pos() + 3 * pointsList[i]->pos() - d2;
		c1 = d1;
		c0 = pointsList[i-1]->pos();

		for(int j=1; j <= hermiteStep; j++)
		{
			u = (double)j/(double)hermiteStep;
			p = c0 + u * (c1 + u * (c2 + u * c3));
			item = new QGraphicsLineItem(QLineF(c0, p));
			item->setPen(QPen(QColor(Qt::green)));
			hermiteLines.append(item);
			addItem(item);
		}
	}

}

void GraphicsScene::drawDeCasteljau()
{
	// Remove the previous lines from the scene
	for(int i=0; i< deCasteljauLines.size(); i++)
		removeItem(deCasteljauLines[i]);

	// Then, delete the lines
	qDeleteAll(deCasteljauLines.begin(), deCasteljauLines.end());
 	deCasteljauLines.clear();

	// Check wheter we have to render the curve, else return
	if(!deCasteljauCurve)
		return;

	// Check wheter we can draw a curve, else return
	if(pointsList.size() < 2)
		return;

	QPointF p1(0.0, 0.0), p2(0.0, 0.0);
	double u = 0.0;
	QGraphicsLineItem * item;
	QVector < QPointF > controlGraph;

	// Fill the control graph
	for(int i=0; i < pointsList.size(); i++)
		controlGraph.append(pointsList[i]->pos());

	p1 = pointsList[0]->pos();

	for(int i=1; i <= deCasteljauStep; i++)
	{
		u = (double)i / (double)deCasteljauStep;

		for(int j = 1; j <= pointsList.size() -1; j++)
			for(int k = 0; k <= pointsList.size() -1 - j; k++)
				controlGraph[k] = (1-u) * controlGraph[k] + u * controlGraph[k+1];

		p2 = controlGraph[0];

		item = new QGraphicsLineItem(QLineF(p1, p2));
		item->setPen(QPen(QColor(Qt::blue)));

		deCasteljauLines.append(item);
		addItem(item);
		p1 = p2;
	}
}

void GraphicsScene::drawBernstein()
{
	// Remove the previous lines from the scene
	for(int i=0; i< bernsteinLines.size(); i++)
		removeItem(bernsteinLines[i]);

	// Then, delete the lines
	qDeleteAll(bernsteinLines.begin(), bernsteinLines.end());
 	bernsteinLines.clear();

	// Check wheter we have to render the curve, else return
	if(!bernsteinCurve)
		return;

	// Check wheter we can draw a curve, else return
	if(pointsList.size() < 2)
		return;

	/*
		Calculate the Tartaglia coefficients
		This algorythm must be horrible, but at least I think it myself :-)
	*/
	QVector < int > tartaglia;
	QVector < int > temp;
	// Initial conditions
	for(int i=0; i < pointsList.size(); i++)
		tartaglia.append(1);

	temp = tartaglia;

	for(int i=2; i < pointsList.size(); i++)
	{
		for(int j=1; j < i; j++)
			tartaglia[j] = temp[j-1] + temp[j];
		temp = tartaglia;
	}

	// Draw the lines
	QPointF p1(0.0, 0.0), p2(0.0, 0.0);
	QGraphicsLineItem * item;
	double u = 1.0 / bernsteinStep;
	double pow1(0.0), pow2(0.0);


	p1 = pointsList[0]->pos();

	for(int j=1; j <= bernsteinStep; j++)
	{
		// p2 = 0
		p2 -= p2;

		for(int k=0; k < tartaglia.size(); k++)
		{
			pow1 = pow(u * j, k);
			pow2 = pow(1 - u * j, tartaglia.size()-1-k);
			p2 = p2 +tartaglia[k] * pow1 * pow2 * pointsList[k]->pos();
		}

		item = new QGraphicsLineItem(QLineF(p1, p2));
		item->setPen(QPen(QColor(Qt::magenta)));

		bernsteinLines.append(item);
		addItem(item);
		p1 = p2;
	}
}

void GraphicsScene::drawBSplines()
{
	// Remove the previous lines from the scene
	for(int i=0; i< bSplineLines.size(); i++)
		removeItem(bSplineLines[i]);

	// Then, delete the lines
	qDeleteAll(bSplineLines.begin(), bSplineLines.end());
 	bSplineLines.clear();

	// Delete the points from the previous calculations
	bSplinePoints.clear();

	// Check wheter we have to render the curve, else return
	if(!bSplinesCurve)
		return;

	// Check wheter we can draw a curve, else return
	if(pointsList.size() < bSplineGrade + 1)
	{
		emit sendSatusBarMessage(tr("The B-Spline algorythm needs at least grade "
		                            "+1 points to work"), 5000);
		return;
	}

	QPointF point;
	int m = pointsList.size() -1;
	double maximumU = (double) (pointsList.size() - bSplineGrade);
	double deltaU = 1.0/bSplineStep;
	double u = deltaU;
	double iBase = 0.0;
	QGraphicsLineItem * item;

	while(u <= maximumU)
	{
		point.rx() = 0.0;
		point.ry() = 0.0;

		for(int i=0; i <= m; i++)
		{
			iBase = splineBase(u, i, bSplineGrade);
			point = point + (iBase * pointsList[i]->pos());
		}

		bSplinePoints.append(point);
		u += deltaU;
	}

	for(int i=0; i < bSplinePoints.size() - 1; i++)
	{
		item = new QGraphicsLineItem(QLineF(bSplinePoints[i], bSplinePoints[i+1]));
		item->setPen(QPen(QColor(Qt::cyan)));

		bSplineLines.append(item);
		addItem(item);
	}
}

double GraphicsScene::splineBase(double u, int i, int grade)
{
	if(grade == 0)
	{
		if((u > node(i)) && (u < node(i+1)))
			return 1.0;
		else
			return 0.0;
	}
	else
	{
		double numerator1 = (u - node(i)) * splineBase(u, i, grade - 1);
		double denominator1 = node(i + grade) - node(i);
		double numerator2 = (node(i + grade + 1) - u) *
		                    splineBase(u, i+1, grade - 1);
		double denominator2 = node(i + grade +1) - node(i + 1);

		double part1 = 0.0;
		double part2 = 0.0;

		if(denominator1 == 0.0)
			part1 = 0.0;
		else
			part1 = numerator1 / denominator1;

		if(denominator2 == 0.0)
			part2 = 0.0;
		else
			part2 = numerator2 / denominator2;

		return part1 + part2;
	}
}

double GraphicsScene::node(int i)
{
	int m = pointsList.size() -1;

	if((0 <= i) && (i <= bSplineGrade))
		return 0;
	else if((bSplineGrade < i) && (i <= m))
		return i - bSplineGrade;
	else
		return m + 1 - bSplineGrade;
}

void GraphicsScene::drawCubicBSpline()
{
	// Remove the previous lines from the scene
	for(int i=0; i< cubicBSplineLines.size(); i++)
		removeItem(cubicBSplineLines[i]);

	// Then, delete the lines
	qDeleteAll(cubicBSplineLines.begin(), cubicBSplineLines.end());
 	cubicBSplineLines.clear();

	// Delete the points
	cubicBSplinePoints.clear();

	// Check wheter we have to render the curve, else return
	if(!cubicBSplineCurve)
		return;

	// Check wheter we can draw a curve, else return
	if(pointsList.size() < 4)
	{
		emit sendSatusBarMessage(tr("The cubic B-Spline algorythm needs at least "
		                            "4 points to work"), 5000);
		return;
	}

	QGraphicsLineItem * item;
	double deltaU = 1.0 / cubicBSplineStep;
	double u = deltaU;
	double a(0.0), b(0.0), c(0.0), d(0.0);
	QPointF temp(0.0, 0.0);

	for(int i=0; i <= pointsList.size() - 4; i++)
	{
		while(u <= 1.0)
		{
			a = (1.0/6.0) * (-u*u*u  + 3*u*u - 3*u + 1);
			b = (1.0/6.0) * (3*u*u*u - 6*u*u + 4);
			c = (1.0/6.0) * (-3*u*u*u + 3*u*u + 3*u +1);
			d = (1.0/6.0) * u*u*u;

			temp = a * pointsList[i]->pos() + b * pointsList[i+1]->pos() +
			       c * pointsList[i+2]->pos() + d * pointsList[i+3]->pos();

			cubicBSplinePoints.append(temp);
			u += deltaU;
		}
		u = deltaU;
	}

	for(int i=0; i < cubicBSplinePoints.size() - 1; i++)
	{
		item = new QGraphicsLineItem(QLineF(cubicBSplinePoints[i],
		                                    cubicBSplinePoints[i+1]));
		item->setPen(QPen(QColor(Qt::darkGreen)));

		cubicBSplineLines.append(item);
		addItem(item);
	}
}


